"""Sqlalchemy2-datatables library, provides serverside implementation for jQuery Datatables"""

__version__ = '0.5.2'

from datatables.base import DTDataCallbacks
from datatables.datatable import DataTable

__all__ = ['DataTable', 'DTDataCallbacks']



from typing import Any
from typing import Callable

from pydantic import BaseModel


class DTKey:
    ROW_ID: str = 'DT_RowId'
    ROW_CLASS: str = 'DT_RowClass'
    ROW_DATA: str = 'DT_RowData'
    ROW_ATTR: str = 'DT_RowAttr'


class DTDataCallbacks:
    get_row_id: Callable[[dict[str, Any]], str] | None = None
    get_row_class: Callable[[dict[str, Any]], str] | None = None
    get_row_data: Callable[[dict[str, Any]], dict[str, Any]] | None = None
    get_row_attr: Callable[[dict[str, Any]], dict[str, Any]] | None = None

    def set_row_id_getter(self, getter: Callable[[dict[str, Any]], str]) -> None:
        """
        Set the lamda that will return the id value of the <tr> node of the datatable row.
        The lambda result will be used to update the result row with DT_RowId key value pair. This will
        set the id attribute of the <tr> node.
        :param: Callable - the lambda that has access to the result row (dict key: column name, value: column value)
                to produce the id value to be added to the <tr> node of the datatable
        """
        self.get_row_id = getter

    def set_row_class_getter(self, getter: Callable[[dict[str, Any]], str]) -> None:
        """
        Set the lamda that will return the class name of the <tr> node of the datatable row.
        The lambda result will be used to update the result row with DT_RowClass key value pair. This will
        add the class to the <tr> node.
        :param: Callable - the lambda that has access to the result row (dict key: column name, value: column value)
                to produce the class name to be added to the <tr> node of the datatable
        """
        self.get_row_class = getter

    def set_row_data_getter(self, getter: Callable[[dict[str, Any]], dict[str, Any]]) -> None:
        """
        Set the lamda that will return the dict of row data to be used by jquery on the client side.
        The lambda result will be used to update the result row with DT_RowData key value pair. This will
        add the data contained in the dict to the jQuery datatables row using the jQuery data() method
        to set the data, which can also then be used for later retrieval (for example on a click event).
        :param: Callable - the lambda that has access to the result row (dict key: column name, value: column value)
                to produce the datatables row data.
        """
        self.get_row_data = getter

    def set_row_attr_getter(self, getter: Callable[[dict[str, Any]], dict[str, str | int | float]]) -> None:
        """
        Set the lamda that will return the dict of attributes that will be added to the <tr> node of
        the datatable row. The dict keys are used as the attribute keys and the values as the corresponding
        attribute values. This is performed using the jQuery param() method.
        The lambda result will be used update the result row with DT_RowData key value pair.
        :param: Callable - the lambda that has access to the result row (dict key: column name, value: column value)
                to produce the dictionary of <tr> attributes.
        """
        self.get_row_attr = getter

    def run(self, row: dict[str, Any]) -> None:
        if self.get_row_id and self.get_row_id(row):
            row[DTKey.ROW_ID] = self.get_row_id(row)

        if self.get_row_class and self.get_row_class(row):
            row[DTKey.ROW_CLASS] = self.get_row_class(row)

        if self.get_row_data and self.get_row_data(row):
            row[DTKey.ROW_DATA] = self.get_row_data(row)

        if self.get_row_attr and self.get_row_attr(row):
            row[DTKey.ROW_ATTR] = self.get_row_attr(row)


class DTColumnOrder(BaseModel):
    """
    Class holding the parameters for one order criterion from the datatable ajax request
    @see https://www.datatables.net/manual/server-side#Sent-parameters

    :param column_index: int - index of the sort column in the table model
    :param is_asc: bool - sort direction for this column, True -> asc, False -> desc
    """

    column_index: int
    is_asc: bool = True


class DTColumn(BaseModel):
    """
    Class holding the parameters for one column from the datatable ajax request
    @see https://www.datatables.net/manual/server-side#Sent-parameters

    :param index: int - index of the column in the table model
    :param data: str - name of the dict key to get the values for this column from the data list object
    :param name: str - name of the column to reference in datatables jQuery javascript
    :param searchable: bool - whether the column should be searchable or not
    :param orderable: bool - whether the column should be orderable or not
    :param search_value: str - the value for the search filter for this column
    :param search_regex: bool - whether the search filter for this column uses regular expressions
    """

    index: int = -1
    data: str = ''
    name: str = ''
    searchable: bool = True
    orderable: bool = True
    search_value: str = ''
    search_regex: bool = False


class DTParams(BaseModel):
    """
    Class holding the query parameters from the datatable ajax request
    @see https://www.datatables.net/manual/server-side#Sent-parameters

    :param draw: int - security id for the ajax request from jQuery datatables
    :param start: int - row to start at, default 0, i.e. offset clause for sql
    :param length: int - number of records returned from the server, default 10, i.e. limit clause for sql
    :param search_value: str | None - global search filter across all searchable columns
    :param search_regex: bool - whether the global search filter uses regular expressions or not
    :param columns: list[DTColumn] - list of column properties
    :param order: list[DTColumnOrder] = list of order criteria
    """

    draw: int = 0
    start: int = 0
    length: int = 10
    search_value: str = ''
    search_regex: bool = False
    columns: list[DTColumn] = []
    order: list[DTColumnOrder] = []



import logging
import random
import re
from typing import Any

from sqlalchemy import FromClause
from sqlalchemy import Result
from sqlalchemy import Select
from sqlalchemy import Subquery
from sqlalchemy import and_
from sqlalchemy import func
from sqlalchemy import or_
from sqlalchemy import select
from sqlalchemy.future import Engine
from sqlalchemy.orm import Session
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy.sql.elements import KeyedColumnElement

from datatables.base import DTColumn
from datatables.base import DTColumnOrder
from datatables.base import DTDataCallbacks
from datatables.base import DTParams


class DataTable:
    """
    Sqlalchemy ORM-compatible data table class.
    See https://www.datatables.net/manual/server-side#API

    :param request_params: dict[str, Any] - the  query parameters sent via the jQuery datatables ajax request
    :param engine: Engine -  sqlalchemy database engine
    :param table: table: FromClause - sqlalchemy FromClause
    :param column_names: list[str] - table column names to display in the datatable, used for projection in sql
    :param callbacks: DTDataCallbacks  - callback that populate  DT_ROW_ID, DT_ROW_CLASS, DT_ROW_ATTR, DT_ROW_DATA
    :attr params: DTParams - parsed request parameters to use for result filtering, projection, sorting and paging
    :attr recordsTotal: int -  the total number of records available in this model/table
    :attr recordsFiltered: int - the number of records for the filtered result (before pagination)
    :attr data: list[dict] - the list of data objects sent to the datatable
    :attr error: str -  if there was an error with data retrieval, this the error message will be sent instead
    """

    callbacks: DTDataCallbacks | None
    engine: Engine
    table: FromClause
    column_names: list[str]
    params: DTParams
    records_total: int = 0
    records_filtered: int = 0
    data: list[dict[str, Any]] = []
    error: str | None = None

    def __init__(
        self,
        request_params: dict[str, Any],
        table: FromClause,
        column_names: list[str],
        engine: Engine,
        callbacks: DTDataCallbacks | None = None,
    ):
        self.table = table
        self.column_names = column_names
        self.engine = engine
        self.callbacks = callbacks
        logging.info(f'initialize DataTable for {self.table}')
        try:
            self.run(request_params)
        except Exception as exc:
            self.error = str(exc)

    @staticmethod
    def _parse_order(request_params: dict[str, Any]) -> list[DTColumnOrder]:
        """Parse the order[index][*] parameters"""
        order: list[DTColumnOrder] = []
        order_pattern = re.compile(r'order\[(.*?)]\[column]')
        order_params: dict[str, Any] = {k: v for k, v in request_params.items() if order_pattern.match(k)}

        for i in range(len(order_params)):
            dt_column_order: DTColumnOrder = DTColumnOrder(
                column_index=request_params[f'order[{i}][column]'],
                is_asc=request_params[f'order[{i}][dir]'] == 'asc',
            )
            order.append(dt_column_order)
        return order

    @staticmethod
    def _parse_columns(request_params: dict[str, Any]) -> list[DTColumn]:
        """Parse the column[index][*] parameters"""
        columns: list[DTColumn] = []
        data_pattern = re.compile(r'columns\[(.*?)]\[data]')
        # Extract only the keys of type columns[i][data] from the params
        data_param: dict[str, Any] = {k: v for k, v in request_params.items() if data_pattern.match(k)}

        for i in range(len(data_param)):
            column: DTColumn = DTColumn(
                index=i,
                data=data_param.get(f'columns[{i}][data]'),
                name=request_params.get(f'columns[{i}][name]'),
                searchable=request_params.get(f'columns[{i}][searchable]') == 'true',
                orderable=request_params.get(f'columns[{i}][orderable]') == 'true',
                search_value=request_params.get(f'columns[{i}][search][value]'),
                search_regex=request_params.get(f'columns[{i}][search][regex]') == 'true',
            )
            columns.append(column)
        return columns

    @staticmethod
    def _parse_params(request_params: dict[str, Any]) -> DTParams:
        """Parse the request (query) parameters"""
        params = DTParams()
        draw: int = int(request_params.get('draw', 0))
        params.draw = random.randint(1, 1000) if draw == 0 else draw
        params.start = int(request_params.get('start', 0))
        params.length = int(request_params.get('length', -1))
        params.search_value = request_params.get('search[value]', '')
        params.search_regex = request_params.get('search[regex]') == 'true'
        params.columns = DataTable._parse_columns(request_params)
        params.order = DataTable._parse_order(request_params)
        logging.info(f'params: {params}')
        return params

    def _get_records_total(self, session: Session) -> int:
        # total record count before filtering
        result: int | None = session.scalar(select(func.count()).select_from(self.table))
        return 0 if result is None else result

    @staticmethod
    def _get_records_filtered(session: Session, subquery: Subquery) -> int:
        result: int | None = session.scalar(select(func.count()).select_from(subquery))
        return 0 if result is None else result

    def _get_table_column_by_index(self, index: int) -> KeyedColumnElement[Any]:
        name: str = self.column_names[index]
        return self.table.columns[name]

    def _add_order_criteria(self, stmt: Select[Any]) -> Select[Any]:
        """Add order by criteria to select statement"""
        for order in self.params.order:
            column: KeyedColumnElement[Any] = self._get_table_column_by_index(order.column_index)
            stmt = stmt.order_by(column.asc()) if order.is_asc else stmt.order_by(column.desc())
        return stmt

    def _add_global_search_criterion(self, stmt: Select[Any]) -> Select[Any]:
        """Add global search filter across all searchable columns to select statement"""
        expressions: list[ColumnElement[Any]] = []
        for dt_col in self.params.columns:
            if dt_col.searchable:
                column: KeyedColumnElement[Any] = self._get_table_column_by_index(dt_col.index)
                col_element: ColumnElement[Any]
                if self.params.search_regex:
                    regex: str = self.params.search_value
                    col_element = column.regexp_match(regex)
                else:
                    col_element = column.like(f'%{self.params.search_value}%')
                expressions.append(col_element)
        return stmt.where(or_(*expressions)) if len(expressions) > 0 else stmt

    def _add_column_search_criteria(self, stmt: Select[Any]) -> Select[Any]:
        """add the individual column filters to select statement"""
        expressions: list[ColumnElement[Any]] = []
        for dt_col in self.params.columns:
            if dt_col.search_value:
                column: KeyedColumnElement[Any] = self._get_table_column_by_index(dt_col.index)
                if self.params.search_regex:
                    col_element = column.regexp_match(dt_col.search_value)
                else:
                    col_element = column.like(f'%{dt_col.search_value}%')
                expressions.append(col_element)
        return stmt.where(and_(*expressions)) if len(expressions) > 0 else stmt

    def _built_select_statement(self) -> Select[Any]:
        stmt: Select[Any] = select().select_from(self.table)
        for name in self.column_names:
            if name not in self.table.columns:
                raise ValueError(f'No column {name} in {self.table}')
            stmt = stmt.add_columns(self.table.columns[name])
        if self.params.search_value:
            stmt = self._add_global_search_criterion(stmt)
        stmt = self._add_column_search_criteria(stmt)
        stmt = self._add_order_criteria(stmt)
        return stmt

    def _get_data(self, session: Session, stmt: Select[Any]) -> list[dict[str, Any]]:
        """Get the data from the database"""
        # adding pagination by page (offset/start) and page size (limit/length)
        stmt = stmt.offset(self.params.start).limit(self.params.length)
        logging.info(f'stmt: {stmt.compile()}')
        result: Result[Any] = session.execute(stmt)
        # create a dictionary that maps the result of the query to a list
        # data: list[dict[str, Any]] = [{k: v for k, v in zip(self.column_names, row)} for row in result.all()]
        data: list[dict[str, Any]] = []
        for row in result.all():
            result_row: dict[str, Any] = {k: v for k, v in zip(self.column_names, row, strict=True)}
            if self.callbacks:
                self.callbacks.run(result_row)
            data.append(result_row)
        return data

    def run(self, request_params: dict[str, Any]) -> None:
        self.params = self._parse_params(request_params)
        with Session(self.engine) as session:
            # get total record count from the table
            self.records_total = self._get_records_total(session)
            # get the select statement with all the search and order criteria
            stmt: Select[Any] = self._built_select_statement()
            # get the filtered record count from the statement that will also produce the data
            self.records_filtered = self._get_records_filtered(session, stmt.subquery())
            # get the filtered records from the database
            self.data = self._get_data(session, stmt)

    def output_result(self) -> dict[str, Any]:
        result: dict[str, Any] = {
            'start': self.params.start,
            'length': self.params.length,
            'draw': self.params.draw,
            'recordsTotal': self.records_total,
            'recordsFiltered': self.records_filtered,
            'data': self.data if self.data else [],
        }
        if self.error:
            result['error'] = self.error
        return result



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['datatables']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.10.4,<2.0.0', 'sqlalchemy>=2.0.0rc2,<3.0.0']

setup_kwargs = {
    'name': 'sqlalchemy2-datatables',
    'version': '0.5.2',
    'description': 'Python Sqlalchemy 2.0 based serverside processing for jQuery datatables.',
    'long_description': '# sqlalchemy2-datatables\n\n[![versions](https://img.shields.io/pypi/pyversions/sqlalchemy2-datatables.svg)](https://github.com/hniedner/sqlalchemy2-datatables)\n[![license](https://img.shields.io/github/license/pydantic/pydantic.svg)](https://github.com/pydantic/pydantic/blob/main/LICENSE)\n[![Python package](https://github.com/coding-doc/sqlalchemy2-datatables/actions/workflows/python-package.yml/badge.svg?branch=main)](https://github.com/coding-doc/sqlalchemy2-datatables/actions/workflows/python-package.yml)\n---\n\n**Source Code**: [https://github.com/coding-doc/sqlalchemy2-datatables](https://github.com/coding-doc/sqlalchemy2-datatables)\n\n---\n### Summary\nsqlalchemy2-datatables is a framework agnostic library providing an SQLAlchemy integration of\njQuery DataTables >= 1.10, and helping you manage server side requests in your application.\n\n### Inspiration\nThis project was inspired by [sqlalchemy-datatables](https://github.com/Pegase745/sqlalchemy-datatables)\ndeveloped by Michel Nemnom aka [Pegase745](https://github.com/Pegase745).\n\n### Motivation\nGiven the sunstantial changes with SQLAlchemy 2.0 most of not all of the SQLAlchemy based datatables serverside\n solution will be outdated soon (labeit currently still supported in SQLAlchemy 1.4). Specifically deprecation of\n sqlalchemy.orm.Query will render those packages obsolete.\n[SQLAlchemy2.0](https://docs.sqlalchemy.org/en/20/).\n\n### Installation\n```shell\npip install sqlalchemy2-datatables\n```\n\n### Examples\nGeneric CRUD style function:\n```python\nfrom typing import Any\nfrom sqlalchemy import Engine\nfrom sqlalchemy import FromClause\n\nfrom datatables import DataTable\nfrom datatables.base import DTDataCallbacks\n\ndef get_datatable_result(\n    params: dict[str, Any],\n    table: FromClause,\n    column_names: list[str],\n    engine: Engine,\n    callbacks: DTDataCallbacks | None,\n) -> dict[str, Any]:\n    """\n    Get database results specifically formatted for a display via jQuery datatables.\n    :param params: dict - request parameters\n    :param table: FromClause - the sqlalchemy from clause\n    :param column_names - List of column names reflecting the table columns in the desired order\n    :param engine: Engine -  the sqlalchemy engine\n    :param callbacks - datatables callbacks to populate jQuery datatables DT_* attributes\n    :return dict with DataTable output for the jQuery datatables in the frontend view\n    """\n    datatable: DataTable = DataTable(\n        request_params=params,\n        table=table,\n        column_names=column_names,\n        engine=engine,\n        callbacks=callbacks\n    )\n    return datatable.output_result()\n```\nThe output dictionary that can be serialized and returned to jQuery datatables.\n```python\n{\n    "start": 0,\n    "length": 5,\n    "draw": 1,\n    "recordsTotal": 1000,\n    "recordsFiltered": 1000,\n    "data": [\n        {\n            "id": 1,\n            "col1": "value",\n            "col2": "value",\n            "col3": "value",\n            "col4": "value",\n        },\n        {\n            "id": 2,\n            "col1": "value",\n            "col2": "value",\n            "col3": "value",\n            "col4": "value",\n        },\n        {\n            "id": 3,\n            "col1": "value",\n            "col2": "value",\n            "col3": "value",\n            "col4": "value",\n        },\n        {\n            "id": 4,\n            "col1": "value",\n            "col2": "value",\n            "col3": "value",\n            "col4": "value",\n        },\n        {\n            "id": 5,\n            "col1": "value",\n            "col2": "value",\n            "col3": "value",\n            "col4": "value",\n        },\n    ],\n}\n```\n',
    'author': 'R. Hannes Niedner',
    'author_email': 'hniedner@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
